@@ -1,7 +1,7 @@
 (********************************************************************)
 (*                                                                  *)
 (*  The Why3 Verification Platform   /   The Why3 Development Team  *)
-(*  Copyright 2010-2020   --   Inria - CNRS - Paris-Sud University  *)
+(*  Copyright 2010-2021 --  Inria - CNRS - Paris-Saclay University  *)
 (*                                                                  *)
 (*  This software is distributed under the terms of the GNU Lesser  *)
 (*  General Public License version 2.1, with the special exception  *)
@@ -25,20 +25,6 @@ open Ity
 open Ty
 open Term
 
-let clean_name fname =
-  (* TODO: replace with Filename.remove_extension
-   * after migration to OCaml 4.04+ *)
-  let remove_extension s =
-    try Filename.chop_extension s with Invalid_argument _ -> s in
-  let f = Filename.basename fname in (remove_extension f)
-
-let module_name ?fname path t =
-  let fname = match fname, path with
-    | None, "why3"::_ -> "why3"
-    | None, _   -> String.concat "__" path
-    | Some f, _ -> clean_name f in
-  fname ^ "__" ^ t
-
 let known_decls : Mltree.decl Hid.t = Hid.create 17
 
 (** Translation from Mlw to ML *)
@@ -69,19 +55,6 @@ module Translate = struct
         if List.length sl < List.length ml then sl (* FIXME ? much likely... *)
         else List.rev (List.fold_left2 add_ity [] ml sl)
 
-  (* types *)
-  let rec type_ ty =
-    match ty.ty_node with
-    | Tyvar tvs ->
-        ML.Tvar tvs
-    | Tyapp (ts, tyl) when is_ts_tuple ts ->
-        ML.Ttuple (List.map type_ tyl)
-    | Tyapp (ts, tyl) ->
-        ML.Tapp (ts.ts_name, List.map type_ tyl)
-
-  let vsty vs =
-    vs.vs_name, type_ vs.vs_ty
-
   let rec filter_ghost_params p def = function
     | [] -> []
     | pv :: l ->
@@ -127,25 +100,6 @@ module Translate = struct
 
   let pv_name pv = pv.pv_vs.vs_name
 
-  (* individual types *)
-  let mlty_of_ity mask t =
-    let rec loop t = match t.ity_node with
-      | _ when mask_equal mask MaskGhost -> ML.tunit
-      | Ityvar tvs ->
-          ML.Tvar tvs
-      | Ityapp ({its_ts = ts}, itl, _) when is_ts_tuple ts ->
-          let itl = visible_of_mask mask itl in
-          ML.Ttuple (List.map loop itl)
-      | Ityapp ({its_ts = ts}, itl, _) ->
-          ML.Tapp (ts.ts_name, List.map loop itl)
-      | Ityreg {reg_its = its; reg_args = args} ->
-          let args = List.map loop args in
-          ML.Tapp (its.its_ts.ts_name, args) in
-    loop t
-
-  let pvty pv =
-    if pv.pv_ghost then ML.mk_var (pv_name pv) ML.tunit true
-    else let (vs, vs_ty) = vsty pv.pv_vs in ML.mk_var vs vs_ty false
 
   let for_direction = function
     | To -> ML.To
@@ -159,7 +113,7 @@ module Translate = struct
         List.exists is_constructor its
     | _ -> false
 
-  let is_singleton_imutable itd =
+  let is_singleton_immutable itd =
     let not_g e = not (rs_ghost e) in
     let pjl = itd.itd_fields in
     let mfields = itd.itd_its.its_mfields in
@@ -180,7 +134,7 @@ module Translate = struct
     | _ -> None
 
   let is_optimizable_record_itd itd =
-    not itd.itd_its.its_private && is_singleton_imutable itd
+    not itd.itd_its.its_private && is_singleton_immutable itd
 
   let is_optimizable_record_rs info rs =
     Opt.fold (fun _ -> is_optimizable_record_itd) false (get_record_itd info rs)
@@ -192,6 +146,29 @@ module Translate = struct
   let is_empty_record info rs =
     Opt.fold (fun _ -> is_empty_record_itd) false (get_record_itd info rs)
 
+  (* individual types *)
+  let mlty_of_ity mask t =
+    let rec loop t = match t.ity_node with
+      | _ when mask_equal mask MaskGhost -> ML.tunit
+      | Ityvar tvs ->
+          ML.Tvar tvs
+      | Ityapp ({its_ts = ts}, itl, _) when is_ts_tuple ts ->
+          let itl = visible_of_mask mask itl in
+          ML.Ttuple (List.map loop itl)
+      | Ityapp (its, [t1; t2], []) when its_equal its its_func ->
+          ML.t_arrow (loop t1) (loop t2)
+      | Ityapp (its, itl, _) ->
+          ML.Tapp (its.its_ts.ts_name, List.map loop itl)
+      | Ityreg {reg_its = its; reg_args = args} ->
+          let args = List.map loop args in
+          ML.Tapp (its.its_ts.ts_name, args) in
+    loop t
+
+  let pvty pv =
+    if pv.pv_ghost then ML.mk_var (pv_name pv) ML.tunit true
+    else
+      ML.mk_var (pv_name pv) (mlty_of_ity MaskVisible pv.pv_ity) false
+
   let mk_eta_expansion rs pvl ({cty_args = ca; cty_effect = ce} as c) =
     (* FIXME : effects and types of the expression in this situation *)
     let rs = Hrs.find_def ht_rs rs rs in
@@ -202,12 +179,16 @@ module Translate = struct
       filter_ghost_params pv_not_ghost def ca in
     let args =
       let def pv = ML.mk_expr (ML.Evar pv) (ML.I pv.pv_ity) mv
-        eff_empty Sattr.empty in
+                     (mlty_of_ity (mask_of_pv pv) pv.pv_ity)
+                     eff_empty Sattr.empty in
       let args = filter_ghost_params pv_not_ghost def pvl in
-      let extra_args = List.map def ca in args @ extra_args in
-    let eapp = ML.mk_expr (ML.Eapp (rs, args)) (ML.C c) mv
-      ce Sattr.empty in
-    ML.mk_expr (ML.Efun (args_f, eapp)) (ML.C c) mv ce Sattr.empty
+      let extra_args = List.map def ca in
+      args @ extra_args in
+    let rty = mlty_of_ity mv c.cty_result in
+    let eapp = ML.mk_expr (ML.Eapp (rs, args, false)) (ML.C c) mv
+                 rty ce Sattr.empty in
+    let mlty = ML.t_fun args_f rty in
+    ML.mk_expr (ML.Efun (args_f, eapp)) (ML.C c) mv mlty ce Sattr.empty
 
   (* function arguments *)
   let filter_params args =
@@ -220,21 +201,26 @@ module Translate = struct
 
   let filter_params_cty p def pvl cty_args =
     let rec loop = function
-      | [], _ -> []
+      | [], rem -> [], List.map def (List.filter p rem)
       | pv :: l1, arg :: l2 ->
-          if p pv && p arg then def pv :: loop (l1, l2) else loop (l1, l2)
+         if p pv && p arg then
+           let applied, remaining = loop (l1, l2) in
+           def pv :: applied, remaining
+         else loop (l1, l2)
       | _ -> assert false
     in loop (pvl, cty_args)
 
   let app pvl cty_args f_zero =
     let def pv = ML.mk_expr (ML.Evar pv) (ML.I pv.pv_ity) MaskVisible
-      eff_empty Sattr.empty in
-    let args = filter_params_cty pv_not_ghost def pvl cty_args in
-    f_zero args
+                   (mlty_of_ity MaskVisible pv.pv_ity)
+                   eff_empty Sattr.empty in
+    let applied, remaining = filter_params_cty pv_not_ghost def pvl cty_args in
+    (f_zero applied, remaining)
 
   (* build the set of type variables from functions arguments *)
   let rec add_tvar acc = function
     | ML.Tvar tv -> Stv.add tv acc
+    | ML.Tarrow (t1, t2) -> add_tvar (add_tvar acc t1) t2
     | ML.Tapp (_, tyl) | ML.Ttuple tyl ->
         List.fold_left add_tvar acc tyl
 
@@ -271,11 +257,11 @@ module Translate = struct
         let c = match c with
           | Constant.ConstInt _ | Constant.ConstStr _ -> c
           | _ -> assert false in
-        ML.e_const c (ML.I e.e_ity) mask eff attrs
+        ML.e_const c (ML.I e.e_ity) mask (mlty_of_ity mask e.e_ity) eff attrs
     | Evar pv ->
         Debug.dprintf debug_compile "compiling variable %a@." print_pv pv;
         assert (mask = MaskVisible);
-        ML.e_var pv (ML.I e.e_ity) mask eff attrs
+        ML.e_var pv (ML.I e.e_ity) mask (mlty_of_ity mask e.e_ity) eff attrs
     | Elet (LDvar (_, e1), e2) when e_ghost e1 ->
         expr info svar mask e2
     | Elet (LDvar (_, e1), e2) when e_ghost e2 ->
@@ -284,13 +270,17 @@ module Translate = struct
     | Elet (LDvar (pv, e1), e2)
       when pv.pv_ghost || not (Mpv.mem pv e2.e_effect.eff_reads) ->
         if eff_pure e1.e_effect then expr info svar mask e2
-        else let e1 = ML.e_ignore e1.e_ity (expr info svar MaskGhost e1) in
-          ML.e_seq e1 (expr info svar mask e2) (ML.I e.e_ity) mask eff attrs
+        else
+          let e1 = (expr info svar MaskGhost e1) in
+          let e1 = ML.e_ignore e1 in
+          let e2 = expr info svar mask e2 in
+          ML.e_seq e1 e2 (ML.I e.e_ity) mask e2.ML.e_mlty eff attrs
     | Elet (LDvar (pv, e1), e2) ->
         Debug.dprintf debug_compile "compiling local definition of %s@."
           (pv_name pv).id_string;
         let ld = ML.var_defn pv (expr info svar MaskVisible e1) in
-        ML.e_let ld (expr info svar mask e2) (ML.I e.e_ity) mask eff attrs
+        let e2 = expr info svar mask e2 in
+        ML.e_let ld e2 (ML.I e.e_ity) mask e2.ML.e_mlty eff attrs
     | Elet (LDsym (rs, _), ein) when rs_ghost rs ->
         expr info svar mask ein
     | Elet (LDsym (_, {c_node = Cpur (_, _); _}), _) ->
@@ -303,9 +293,11 @@ module Translate = struct
         let new_svar = new_svar args res svar in
         let e_let = expr info svar cty.cty_mask ef in
         let ld = ML.sym_defn rs new_svar res args e_let in
-        ML.e_let ld (expr info svar mask ein) (ML.I e.e_ity) mask eff attrs
+        let ein = expr info svar mask ein in
+        ML.e_let ld ein (ML.I e.e_ity) mask (ein.ML.e_mlty) eff attrs
     | Elet (LDsym (rs, {c_node = Capp (rs_app, pvl); c_cty = cty}), ein)
-      when isconstructor info rs_app -> (* partial application of constructor *)
+         when isconstructor info rs_app -> (* partial application of constructor *)
+        Debug.dprintf debug_compile "partial application of constructor@.";
         let eta_app = mk_eta_expansion rs_app pvl cty in
         let mk_func pv f = ity_func pv.pv_ity f in
         let func = List.fold_right mk_func cty.cty_args cty.cty_result in
@@ -314,22 +306,32 @@ module Translate = struct
         let new_svar = new_svar args res svar in
         let ld = ML.sym_defn rs new_svar res [] eta_app in
         let ein = expr info svar mask ein in
-        ML.e_let ld ein (ML.I e.e_ity) mask eff attrs
+        ML.e_let ld ein (ML.I e.e_ity) mask (ein.ML.e_mlty) eff attrs
     | Elet (LDsym (rsf, {c_node = Capp (rs_app, pvl); c_cty = cty}), ein) ->
-        (* partial application *) (* FIXME -> zero arguments functions *)
+        (* partial application *)
         Debug.dprintf debug_compile "compiling partial application of %s@."
           rsf.rs_name.id_string;
         let cmk = cty.cty_mask in
         let ceff = cty.cty_effect in
-        let pvl = app pvl rs_app.rs_cty.cty_args (fun x -> x) in
         let rs_app = Hrs.find_def ht_rs rs_app rs_app in
-        let eapp = ML.e_app rs_app pvl (ML.C cty) cmk ceff Sattr.empty in
-        let args = params cty.cty_args in
+        let (pvl, remaining) = app pvl rs_app.rs_cty.cty_args (fun x -> x) in
         let res = mlty_of_ity cty.cty_mask cty.cty_result in
-        let new_svar = new_svar args res svar in
-        let ld = ML.sym_defn rsf new_svar res args eapp in
+        let fargs =
+          match remaining, rs_app.rs_logic with
+          | [], RLnone ->
+             (* application is partial, so some ghost arguments
+                must still be missing *)
+             [ML.mk_var_unit]
+          | _, _ -> [] in
+        let appty =
+          List.fold_right
+            (fun pv rty -> ML.t_arrow pv.ML.e_mlty rty)
+            remaining res in
+        let eapp = ML.e_app rs_app pvl true
+                     (ML.C cty) cmk appty ceff Sattr.empty in
+        let ld = ML.sym_defn rsf Stv.empty appty fargs eapp in
         let ein = expr info svar mask ein in
-        ML.e_let ld ein (ML.I e.e_ity) mask eff attrs
+        ML.e_let ld ein (ML.I e.e_ity) mask ein.ML.e_mlty eff attrs
     | Elet (LDsym (_, {c_node = Cany; _}), _) -> let loc = e.e_loc in
         Loc.errorm ?loc "This expression cannot be extracted"
     | Elet (LDrec rdefl, ein) ->
@@ -350,20 +352,25 @@ module Translate = struct
         if rdefl <> [] then
           let ein = expr info svar mask ein in
           let ml_letrec = ML.Elet (ML.Lrec rdefl, ein) in
-          ML.mk_expr ml_letrec (ML.I e.e_ity) mask e.e_effect attrs
+          ML.mk_expr ml_letrec (ML.I e.e_ity) mask ein.ML.e_mlty
+            e.e_effect attrs
         else expr info svar mask ein
     | Eexec ({c_node = Capp (rs, [])}, _)  when is_rs_tuple rs ->
         ML.e_unit
     | Eexec ({c_node = Capp (rs, pvl)}, _) when is_rs_tuple rs ->
         let pvl = pv_list_of_mask pvl mask in
+        let mltyl =
+          List.map (fun pv -> mlty_of_ity MaskVisible pv.pv_ity) pvl in
         let res_ity = ity_tuple (List.map (fun v -> v.pv_ity) pvl) in
-        let pvl = ML.var_list_of_pv_list pvl in
-        ML.e_app rs pvl (ML.I res_ity) mask eff attrs
+        let pvl = ML.var_list_of_pv_list pvl mltyl in
+        ML.e_app rs pvl false
+          (ML.I res_ity) mask (mlty_of_ity mask res_ity) eff attrs
     | Eexec ({c_node = Capp (rs, _)}, _) when is_empty_record info rs ->
         ML.e_unit
     | Eexec ({c_node = Capp (rs, pvl); c_cty = cty}, _)
       when isconstructor info rs && cty.cty_args <> [] ->
-        (* partial application of constructors *)
+       (* partial application of constructors *)
+        Debug.dprintf debug_compile "partial application of constructor@.";
         mk_eta_expansion rs pvl cty
     | Eexec ({c_node = Capp (rs, pvl); c_cty = cty}, _) ->
         Debug.dprintf debug_compile "compiling application of %s@."
@@ -379,26 +386,36 @@ module Translate = struct
               (* FIXME: ideally this should be done in ocaml_printer *)
               add_unit
           | _ -> id_f in
-        let pvl = app pvl rs.rs_cty.cty_args f_zero in
+        let (pvl,rem) = app pvl rs.rs_cty.cty_args f_zero in
         let eta_exp_pj is_optimizable =
           Debug.dprintf debug_compile "record projection@.";
           let params = filter_params cty.cty_args in
           let args = cty.cty_args in
-          let app_args = app args args (fun x -> x) in
+          let (app_args,rem) = app args args (fun x -> x) in
+          assert (rem = []);
           (* create the identity function *)
-          let ml_app = if is_optimizable then
+          let ml_app =
+            if is_optimizable then
               match app_args with [a] -> a | _ -> assert false
-            else let ity_res = (ML.I rs.rs_cty.cty_result) in
-              ML.e_app rs app_args ity_res MaskVisible eff_empty attrs in
+            else
+              let ity_res = (ML.I rs.rs_cty.cty_result) in
+              let mlty_res = mlty_of_ity MaskVisible rs.rs_cty.cty_result in
+              ML.e_app rs app_args false ity_res MaskVisible
+                mlty_res eff_empty attrs
+          in
           let ity = ML.C cty in
           let attrs = Sattr.empty in
-          ML.e_fun params ml_app ity cty.cty_mask eff_empty attrs in
+          let fun_ty = ML.t_fun params ml_app.ML.e_mlty in
+          ML.e_fun params ml_app ity cty.cty_mask fun_ty
+            eff_empty attrs in
         begin match pvl with
           | [pv_expr] when is_optimizable_record_rs info rs -> pv_expr
           | []        when is_optimizable_record_rs info rs ->
               eta_exp_pj true
           | [] when rs.rs_field <> None -> eta_exp_pj false
-          | _ -> ML.e_app rs pvl (ML.I e.e_ity) mask eff attrs end
+          | _ ->
+             let mlty = mlty_of_ity mask e.e_ity in
+             ML.e_app rs pvl (rem <> []) (ML.I e.e_ity) mask mlty eff attrs end
     | Eexec ({c_node = Cfun e; c_cty = {cty_args = []}}, _) ->
         (* abstract block *)
         Debug.dprintf debug_compile "compiling abstract block@.";
@@ -407,13 +424,16 @@ module Translate = struct
         (* is it the case that every argument here is non-ghost ? *)
         Debug.dprintf debug_compile "compiling a lambda expression@.";
         let ef = expr info svar e.e_mask ef in
-        ML.e_fun (params cty.cty_args) ef (ML.I e.e_ity) mask eff attrs
+        let mlty = mlty_of_ity mask e.e_ity in
+        ML.e_fun (params cty.cty_args) ef
+          (ML.I e.e_ity) mask mlty eff attrs
     | Eexec ({c_node = Cpur (_, _); _ }, _) ->
         assert false (* necessarily ghost *)
     | Eexec ({c_node = Cany}, _) -> let loc = e.e_loc in
         Loc.errorm ?loc "This expression cannot be extracted"
     | Eabsurd ->
-        ML.e_absurd (ML.I e.e_ity) mask eff attrs
+        let mlty = mlty_of_ity mask e.e_ity in
+        ML.e_absurd (ML.I e.e_ity) mask mlty eff attrs
     | Eassert _ ->
         ML.e_unit
     | Eif (e1, e2, e3) when e_ghost e1 ->
@@ -424,33 +444,46 @@ module Translate = struct
     | Eif (e1, e2, e3) when e_ghost e3 ->
         let e1 = expr info svar e1.e_mask e1 in
         let e2 = expr info svar mask e2 in
-        ML.e_if e1 e2 ML.e_unit mask eff attrs
+        assert (e2.ML.e_mlty = ML.tunit);
+        ML.e_if e1 e2 ML.e_unit mask ML.tunit eff attrs
     | Eif (e1, e2, e3) when e_ghost e2 ->
         let e1 = expr info svar e1.e_mask e1 in
         let e3 = expr info svar mask e3 in
-        ML.e_if e1 ML.e_unit e3 mask eff attrs
+        assert (e3.ML.e_mlty = ML.tunit);
+        ML.e_if e1 ML.e_unit e3 mask ML.tunit eff attrs
     | Eif (e1, e2, e3) -> Debug.dprintf debug_compile "compiling if block@.";
         let e1 = expr info svar e1.e_mask e1 in
         let e2 = expr info svar mask e2 in
         let e3 = expr info svar mask e3 in
-        ML.e_if e1 e2 e3 mask eff attrs
+        assert (e2.ML.e_mlty = e3.ML.e_mlty);
+        ML.e_if e1 e2 e3 mask e2.ML.e_mlty eff attrs
     | Ewhile (e1, _, _, e2) ->
         Debug.dprintf debug_compile "compiling while block@.";
         let e1 = expr info svar e1.e_mask e1 in
         let e2 = expr info svar e2.e_mask e2 in
-        ML.e_while e1 e2 mask eff attrs
+        assert (e2.ML.e_mlty = ML.tunit);
+        ML.e_while e1 e2 mask ML.tunit eff attrs
     | Efor (pv1, (pv2, dir, pv3), _, _, efor) ->
         Debug.dprintf debug_compile "compiling for block@.";
         let dir = for_direction dir in
         let efor = expr info svar efor.e_mask efor in
-        ML.e_for pv1 pv2 dir pv3 efor mask eff attrs
+        let imlty = mlty_of_ity MaskVisible pv1.pv_ity in
+        assert (efor.ML.e_mlty = ML.tunit);
+        ML.e_for pv1 imlty pv2 dir pv3 efor mask ML.tunit eff attrs
     | Eghost _ | Epure _ ->
         assert false
     | Eassign al ->
         let rm_ghost (_, rs, _) = not (rs_ghost rs) in
         let al = List.filter rm_ghost al in
-        let e_of_var pv = ML.e_var pv (ML.I pv.pv_ity) MaskVisible eff attrs in
-        let al = List.map (fun (pv1, rs, pv2) -> (pv1, rs, e_of_var pv2)) al in
+        let e_of_var pv =
+          ML.e_var pv (ML.I pv.pv_ity) MaskVisible
+            (mlty_of_ity MaskVisible pv.pv_ity) eff attrs in
+        let al =
+          List.map
+            (fun (pv1, rs, pv2) ->
+              (e_of_var pv1, mlty_of_ity MaskVisible pv1.pv_ity, rs,
+               e_of_var pv2)) al
+        in
         ML.e_assign al (ML.I e.e_ity) mask eff attrs
     | Ematch (e1, bl, xl) when e_ghost e1 ->
         assert (Mxs.is_empty xl); (* Expr ensures this for the time being *)
@@ -469,18 +502,25 @@ module Translate = struct
           if e.e_effect.eff_ghost then (xs, pvl, ML.e_unit)
           else (xs, pvl, expr info svar mask e) in
         let xl = List.map mk_xl (Mxs.bindings xl) in
-        ML.e_match e1 bl xl (ML.I e.e_ity) mask eff attrs
-    | Eraise (xs, ex) -> let ex = match expr info svar xs.xs_mask ex with
-        | {ML.e_node = ML.Eblock []} -> None
-        | e -> Some e in
-        ML.mk_expr (ML.Eraise (xs, ex)) (ML.I e.e_ity) mask eff attrs
+        let mlty = mlty_of_ity mask e.e_ity in
+        ML.e_match e1 bl xl (ML.I e.e_ity) mask mlty eff attrs
+    | Eraise (xs, ex) ->
+        let ex = match expr info svar xs.xs_mask ex with
+          | {ML.e_node = ML.Eblock []} -> None
+          | e -> Some e in
+        let mlty = mlty_of_ity mask e.e_ity in
+        ML.mk_expr (ML.Eraise (xs, ex)) (ML.I e.e_ity) mask mlty eff attrs
     | Eexn (xs, e1) ->
-        if mask_ghost e1.e_mask then ML.mk_expr
-          (ML.Eexn (xs, None, ML.e_unit)) (ML.I e.e_ity) mask eff attrs
-        else let e1 = expr info svar xs.xs_mask e1 in
-          let ty = if ity_equal xs.xs_ity ity_unit then None
-            else Some (mlty_of_ity xs.xs_mask xs.xs_ity) in
-        ML.mk_expr (ML.Eexn (xs, ty, e1)) (ML.I e.e_ity) mask eff attrs
+       let mlty = mlty_of_ity mask e.e_ity in
+       if mask_ghost e1.e_mask
+       then
+         ML.mk_expr
+           (ML.Eexn (xs, None, ML.e_unit)) (ML.I e.e_ity) mask mlty eff attrs
+       else
+         let e1 = expr info svar xs.xs_mask e1 in
+         let ty = if ity_equal xs.xs_ity ity_unit then None
+                  else Some (mlty_of_ity xs.xs_mask xs.xs_ity) in
+        ML.mk_expr (ML.Eexn (xs, ty, e1)) (ML.I e.e_ity) mask mlty eff attrs
 
   and ebranch info svar mask ({pp_pat = p; pp_mask = m}, e) =
     (* if the [case] expression is not ghost but there is (at least) one ghost
@@ -495,7 +535,8 @@ module Translate = struct
     let ddata_constructs = (* point-free *)
       List.map (fun ({rs_cty = cty} as rs) ->
         let args = List.filter pv_not_ghost cty.cty_args in
-        (rs.rs_name, List.map (fun {pv_vs = vs} -> type_ vs.vs_ty) args)) in
+        (rs.rs_name,
+         List.map (fun pv-> mlty_of_ity MaskVisible pv.pv_ity) args)) in
     let drecord_fields ({rs_cty = cty} as rs) =
       (List.exists (pv_equal (fd_of_rs rs)) s.its_mfields,
        rs.rs_name,
@@ -514,7 +555,9 @@ module Translate = struct
           let pjl = filter_ghost_params p drecord_fields pjl in
           begin match pjl with
             | [] ->
-                ML.mk_its_defn id args is_private (Some (ML.Dalias ML.tunit))
+                let ty_def = if is_private then None
+                  else Some (ML.Dalias (ML.tunit)) in
+                ML.mk_its_defn id args is_private ty_def
             | [_, _, ty_pj] when is_optimizable_record_itd itd ->
                 ML.mk_its_defn id args is_private (Some (ML.Dalias ty_pj))
             | pjl ->
@@ -583,9 +626,9 @@ module Translate = struct
         let e = expr info svar cty.cty_mask e in
         [ML.Dlet (ML.Lsym (rs, svar, res, args, e))]
     | PDlet (LDsym ({rs_cty = cty} as rs, {c_node = Cfun e})) ->
-        Debug.dprintf debug_compile "compiling function %a@." Expr.print_rs rs;
         let args = params cty.cty_args in
         let res = mlty_of_ity cty.cty_mask cty.cty_result in
+        Debug.dprintf debug_compile "compiling function %a@." Expr.print_rs rs;
         let svar = new_svar args res Stv.empty in
         let e = expr info svar cty.cty_mask e in
         [ML.Dlet (ML.Lsym (rs, svar, res, args, e))]
@@ -596,7 +639,8 @@ module Translate = struct
         let def {rec_fun = e; rec_sym = rs1} =
           let e = match e.c_node with Cfun e -> e | _ -> assert false in
           let args = params rs1.rs_cty.cty_args in
-          let res  = mlty_of_ity rs1.rs_cty.cty_mask rs1.rs_cty.cty_result in
+          let res  = mlty_of_ity rs1.rs_cty.cty_mask
+                       rs1.rs_cty.cty_result in
           let svar =
             let args' = List.map (fun (_, ty, _) -> ty) args in
             let svar  = List.fold_left add_tvar Stv.empty args' in
@@ -659,15 +703,26 @@ module RefreshLetBindings = struct
     let id' = id_clone (pv_name pv) in
     create_pvsymbol id' pv.pv_ity
 
-  let rec expr ((accv, accf) as acc) e =
+  let rec expr ((accv, accf, accx) as acc) e =
     let acc,e = e_map_fold expr acc e in
     let mk e_node = { e with e_node = e_node } in
     match e.e_node with
-    (* collect let-bindings *)
+    (* collect bindings *)
     | Elet (ld, e) ->
        let acc, ld' = ldef acc ld in
        let acc, e' = expr acc e in
        acc, mk (Elet (ld', e'))
+    | Ematch (e, bl, el) ->
+        let acc, bl = Lists.map_fold_left match_pat acc bl in
+        let acc, el = Lists.map_fold_left match_exn acc el in
+        acc, mk (Ematch (e, bl, el))
+    | Eexn (xs, ty, e) ->
+        assert (not (Mid.mem xs.xs_name accx));
+        let id = id_clone xs.xs_name in
+        let xs = create_xsymbol id ~mask:xs.xs_mask xs.xs_ity in
+        let acc = accv, accf, Mid.add xs.xs_name xs accx in
+        let acc, e = expr acc e in
+        acc, mk (Eexn (xs, ty, e))
     (* apply transformation under lambdas *)
     | Efun (vl, e) ->
        assert (List.for_all (fun (id,_,_) -> not (Mid.mem id accv)) vl);
@@ -677,35 +732,27 @@ module RefreshLetBindings = struct
     | Evar pv ->
        let pv' = pvs accv pv in
        acc, mk (Evar pv')
-    | Eapp (rs, el) ->
+    | Eapp (rs, el, p) ->
        let rs' = Mrs.find_def rs rs accf in
-       acc, mk (Eapp (rs', el))
-    | Eassign al ->
-       let al' =
-         List.map
-           (fun (pv, rs, e) ->
-             let pv' = pvs accv pv in
-             assert (not (Mrs.mem rs accf));
-             (pv', rs, e))
-           al in
-       acc, mk (Eassign al')
-    | _ -> acc, e
+       acc, mk (Eapp (rs', el, p))
+    | Eraise (xs, e) ->
+        let xs = Mid.find_def xs xs.xs_name accx in
+        acc, mk (Eraise (xs, e))
+    | Econst _ | Eif _ | Eassign _ | Eblock _ | Ewhile _ | Efor _
+    | Eignore _ | Eabsurd -> acc, e
 
   and pvs accv pv = Mid.find_def pv (pv_name pv) accv
 
-  and ldef ((accv, accf) as acc) ld =
+  and ldef ((accv, accf, accx) as acc) ld =
     match ld with
     | Lvar (pv, e) ->
-      let id = Translate.pv_name pv in
-      assert (not (Mid.mem id accv));
-      let pv' = clone_pv pv in
-      let acc = (Mid.add id pv' accv, accf) in
-      let acc', e' = expr acc e in
-      acc', Lvar (pv', e')
+        let (acc, pv) = refresh_pv acc pv in
+        let acc, e = expr acc e in
+        acc, Lvar (pv, e)
     | Lsym (rs, tv, rty, vl, e) ->
        assert (List.for_all (fun (id,_,_) -> not (Mid.mem id accv)) vl);
        let rs' = clone_rs rs in
-       let acc = (accv, Mrs.add rs rs' accf) in
+       let acc = (accv, Mrs.add rs rs' accf, accx) in
        let acc', e' = expr acc e in
        acc', Lsym (rs, tv, rty, vl, e')
     | Lany _ -> acc, ld
@@ -717,6 +764,7 @@ module RefreshLetBindings = struct
              let rs' = clone_rs rs in
              Mrs.add rs rs' acc, { rd with rec_sym = rs' })
            accf rl in
+       let acc = accv, accf, accx in
        let acc, rl =
          Lists.map_fold_left
            (fun acc rd ->
@@ -724,10 +772,49 @@ module RefreshLetBindings = struct
                        (fun (id,_,_) -> not (Mid.mem id accv))
                        rd.rec_args);
              let acc, e = expr acc rd.rec_exp in
-             acc, { rd with rec_exp = e }) (accv, accf) rl in
+             acc, { rd with rec_exp = e }) acc rl in
        acc, Lrec rl
 
-  let expr e = let _, e' = expr (Mid.empty, Mrs.empty) e in e'
+  and refresh_pv (accv, accf, accx) pv =
+    let id = Translate.pv_name pv in
+    assert (not (Mid.mem id accv));
+    let pv = clone_pv pv in
+    let acc = Mid.add id pv accv, accf, accx in
+    acc, pv
+
+  and match_pat acc (p, e) =
+    let rec aux acc = function
+      | Pwild ->
+          acc, Pwild
+      | Pvar vs ->
+          let pv = restore_pv vs in
+          let acc, pv = refresh_pv acc pv in
+          acc, Pvar pv.pv_vs
+      | Papp (ls, pl) ->
+          let acc, pl = Lists.map_fold_left aux acc pl in
+          acc, Papp (ls, pl)
+      | Ptuple pl ->
+          let acc, pl = Lists.map_fold_left aux acc pl in
+          acc, Ptuple pl
+      | Por (p1, p2) ->
+          let acc, p1 = aux acc p1 in
+          let acc, p2 = aux acc p2 in
+          acc, Por (p1, p2)
+      | Pas (pat, vs) ->
+          let pv = restore_pv vs in
+          let acc, pv = refresh_pv acc pv in
+          acc, Pas (pat, pv.pv_vs) in
+    let acc, p = aux acc p in
+    let acc, e = expr acc e in
+    acc, (p, e)
+
+  and match_exn ((_accv, _accf, accx) as acc) (xs, pl, e) =
+    let xs = Mid.find_def xs xs.xs_name accx in
+    let acc, pl = Lists.map_fold_left refresh_pv acc pl in
+    let acc, e = expr acc e in
+    acc, (xs, pl, e)
+
+  let expr e = let _, e' = expr (Mid.empty, Mrs.empty, Mid.empty) e in e'
 
 end
 
@@ -746,7 +833,7 @@ module InlineFunctionCalls = struct
     let mk e_node = { e with e_node = e_node } in
     match e.e_node with
     | Evar v -> mk (Evar (pv subst v))
-    | Eapp (rs, el) when Sattr.mem inline_attr rs.rs_name.id_attrs ->
+    | Eapp (rs, el, _) when Sattr.mem inline_attr rs.rs_name.id_attrs ->
        let fname = rs.rs_name in
        Debug.dprintf debug_compile "inlining call to %s@." fname.id_string;
        let decl = Hid.find known_decls fname in
@@ -777,15 +864,9 @@ module InlineFunctionCalls = struct
     | Efun (vl, e) ->
        List.iter (fun (id, _ty, _gh) -> assert (not (Mid.mem id subst))) vl;
        mk (Efun (vl, expr subst e))
-    | Eassign al ->
-       let assign (v, rs, e) =
-         let pv' = pv subst v in
-         (pv', rs, e) in
-       let al' = List.map assign al in
-       mk (Eassign al')
-    | Efor (i, st, dir, en, e) ->
+    | Efor (i, ty, st, dir, en, e) ->
        assert (not (Mid.mem (pv_name i) subst));
-       mk (Efor (i, pv subst st, dir, pv subst en, e))
+       mk (Efor (i, ty, pv subst st, dir, pv subst en, e))
     | _ -> e
 
   and pv subst v = try pv subst (Mid.find (pv_name v) subst) with Not_found -> v
@@ -849,7 +930,7 @@ module InlineProxyVars = struct
            with zero arguments inside a [Efun]. *)
         let p _k e = match e.e_node with
           | Econst _ -> true
-          | Eapp (rs, []) when Translate.isconstructor info rs -> true
+          | Eapp (rs, [], _) when Translate.isconstructor info rs -> true
           | _ -> false in
         let restrict_subst = Mpv.filter p subst in
         (* We begin the inlining of proxy variables in an [Efun] with a
@@ -857,10 +938,8 @@ module InlineProxyVars = struct
            undiserable captures inside the [Efun] expression. *)
         let (spv,occ), e = expr info restrict_subst (vars,occ) e in
         (spv,occ), mk (Efun (vl, e))
-    | Efor (pv1, pv2, _, pv3, _) ->
+    | Efor (pv1, _, pv2, _, pv3, _) ->
        (vars, Spv.add pv1 (Spv.add pv2 (Spv.add pv3 occ))), e
-    | Eassign al ->
-       (vars, List.fold_left (fun occ (pv, _, _) -> Spv.add pv occ) occ al), e
     | _ -> (vars, occ), e
 
   and let_def info subst (vars,occ) ld =
@@ -885,7 +964,7 @@ module InlineProxyVars = struct
 
 end
 
-module InlineTrivialLets = struct
+module ExprSimplifications = struct
 
   open Mltree
 
@@ -893,6 +972,13 @@ module InlineTrivialLets = struct
     let e = e_map expr e in
     match e.e_node with
     | Elet (Lvar (pv1, e'), { e_node = Evar pv2 }) when pv_equal pv1 pv2 -> e'
+    | Ematch (e1, [Pwild, e2], []) ->
+       mk_expr (Eblock [e_ignore e1; e2]) e.e_ity MaskVisible
+         e.e_mlty e.e_effect e.e_attrs
+    | Ematch (e1, [Pvar v, e2], []) ->
+       let pv = restore_pv v in
+       e_let (Lvar (pv, e1)) e2 e.e_ity MaskVisible
+         e.e_mlty e.e_effect e.e_attrs
     | _ -> e
 
   let let_def ld = ld_map expr ld
@@ -918,6 +1004,6 @@ module Transform = struct
   let module_ m =
     let m = InlineFunctionCalls.module_ m in
     let m = InlineProxyVars.module_ m in
-    InlineTrivialLets.module_ m
+    ExprSimplifications.module_ m
 
 end
