@@ -1,7 +1,7 @@
 (********************************************************************)
 (*                                                                  *)
 (*  The Why3 Verification Platform   /   The Why3 Development Team  *)
-(*  Copyright 2010-2020   --   Inria - CNRS - Paris-Sud University  *)
+(*  Copyright 2010-2021 --  Inria - CNRS - Paris-Saclay University  *)
 (*                                                                  *)
 (*  This software is distributed under the terms of the GNU Lesser  *)
 (*  General Public License version 2.1, with the special exception  *)
@@ -179,7 +179,9 @@ let print_id_real info fmt id =
     let th = get_th_name th.Theory.th_name in
     let id = id_unique ipr id in
     if path = "" then fprintf fmt "%s.%s" th id
-    else fprintf fmt "%s@@%s.%s" path th id
+    else
+      (* `%@` prints a single `@` *)
+      fprintf fmt "%s%@%s.%s" path th id
   with Not_found ->
     print_id fmt id
 
