@@ -447,6 +447,7 @@ let timeout_handler () =
   (* examine all the prover tasks in progress *)
   (* When no tasks are there, probably no tasks were scheduled and the server
      was not launched so getting results could fail. *)
+  let need_update = ref false in
   if Hashtbl.length prover_tasks_in_progress != 0 then begin
     let results = Call_provers.get_new_results ~blocking:S.blocking in
     List.iter (fun (call, prover_update) ->
@@ -476,7 +477,8 @@ let timeout_handler () =
             if ptp.tp_started then decr number_of_running_provers;
             (* inform the callback *)
             ptp.tp_callback (InternalFailure (exn))
-        end
+        end;
+        need_update := !need_update || prover_update != Call_provers.NoUpdates
       | exception Not_found -> ()
         (* We probably received ProverStarted after ProverFinished,
            because what is sent to and received from the server is
@@ -504,7 +506,7 @@ let timeout_handler () =
     done;
     Queue.transfer q prover_tasks_edited;
   end;
-  update_observer ();
+  if !need_update then update_observer ();
   true
 
 
@@ -520,12 +522,13 @@ let idle_handler () =
           S.multiplier * !session_max_tasks)
       then
         let spa = Queue.pop scheduled_proof_attempts in
-        try build_prover_call spa
+        try
+          build_prover_call spa;
+          update_observer ()
         with e when not (Debug.test_flag Debug.stack_trace) ->
           spa.spa_callback (InternalFailure e)
     with Queue.Empty -> idle_handler_running := false
   end;
-  update_observer ();
   !idle_handler_running
 
 let interrupt () =
@@ -763,25 +766,26 @@ let schedule_transformation c id name args ~callback ~notification =
     callback s
   in
   let apply_trans () =
-    begin
+    let status =
       try
         let subtasks =
           apply_trans_to_goal ~allow_no_effect:false
                               c.controller_session c.controller_env name args id
         in
         let tid = graft_transf c.controller_session id name args subtasks in
-        callback (TSdone tid)
+        TSdone tid
       with
       | NoProgress ->
           (* if result is same as input task, consider it as a failure *)
-          callback (TSfailed (id, NoProgress))
+          TSfailed (id, NoProgress)
       | e when not (is_fatal e) ->
-          callback (TSfailed (id, e))
+          TSfailed (id, e)
       | e when not (Debug.test_flag Debug.stack_trace) ->
           (* "@[Exception raised in Session_itp.apply_trans_to_goal %s:@ %a@.@]"
           name Exn_printer.exn_printer e; TODO *)
-          callback (TSfatal (id, e))
-    end;
+          TSfatal (id, e)
+    in
+    callback status;
     false
   in
   if Session_itp.is_detached c.controller_session (APn id) then
@@ -812,9 +816,9 @@ let run_strategy_on_goal
            | Done { Call_provers.pr_answer = Call_provers.Valid } ->
               (* proof succeeded, nothing more to do *)
               callback STShalt
-           | Interrupted | InternalFailure _ ->
+           | Interrupted ->
               callback STShalt
-           | Done _ ->
+           | Done _ | InternalFailure _ ->
               (* proof did not succeed, goto to next step *)
               callback (STSgoto (g,pc+1));
               let run_next () = exec_strategy (pc+1) strat g; false in
