@@ -1,7 +1,7 @@
 (********************************************************************)
 (*                                                                  *)
 (*  The Why3 Verification Platform   /   The Why3 Development Team  *)
-(*  Copyright 2010-2020   --   Inria - CNRS - Paris-Sud University  *)
+(*  Copyright 2010-2021 --  Inria - CNRS - Paris-Saclay University  *)
 (*                                                                  *)
 (*  This software is distributed under the terms of the GNU Lesser  *)
 (*  General Public License version 2.1, with the special exception  *)
@@ -70,3 +70,46 @@ val ffalse : 'a -> bool
 
 val ttrue : 'a -> bool
 (** [ttrue] constant function [true] *)
+
+(** {3 Lexical comparison using projections}
+
+    For example to lexically sort a list [l] of pairs [(int * string) list]:
+
+      [cmp [cmptr fst Int.compare; cmptr snd String.compare] l] *)
+
+type 'a cmptr
+(** A comparator for values of type ['a] **)
+
+type 'a compare = 'a -> 'a -> int
+
+val cmptr : ('a -> 'b) -> 'b compare -> 'a cmptr
+(** Create a comparator by a projection and a comparison function between projected values *)
+
+val cmptr_direct : 'a compare -> 'a cmptr
+
+val cmp : 'a cmptr list -> 'a compare
+(** Create a comparison function using lexical order defined by a list of comparators *)
+
+val cmp_lists : 'a cmptr list -> 'a list compare
+(** Create a comparison function for lists using lexical order defined by a list of comparators *)
+
+(** {3 ANSI terminal colors} *)
+
+val terminal_has_color : bool
+(** Indicates if standard output supports ANSI terminal color codes (i.e. that the
+   ["TERM"] environment variables is set, and not to ["dump"], and that standard output is
+   a terminal. *)
+
+val ansi_color_tags : Format.formatter_tag_functions
+(** Functions to interpret tags as ANSI terminal color codes. The format of the tag is
+   [[bold] [<color>] [on <bg-color>]].
+
+    Possible colors are [black], [red], [green], [yellow], [blue], [magenta], [cyan], and
+   [white].
+
+    Valid formatting tags are for example ["@{<red>red text@}"], ["@{<bold>bold text@}"],
+   ["@{<on green>text on green background@}"], or ["@{<bold red on green>unreadable@}"].
+   *)
+
+(** Check if a non-empty string contains only characters [a-zA-Z0-9_-] *)
+val is_sexp_simple_token : string -> bool
